from transformers import GPT2LMHeadModel
import torch.nn as nn
from adapter import Adapter

    

class BapnaAdapter(Adapter):
    """
    A variant of the Houlsby Adapter. 
    For more information, please refer to the following paper:
    https://aclanthology.org/D19-1165.pdf
    """
    def __init__(self, size = 6, model_dim = 768):
        super(BapnaAdapter, self).__init__(size=size, model_dim=model_dim)

        self.adapter_block = nn.Sequential(
            nn.LayerNorm(model_dim),
            nn.Linear(model_dim, size),
            nn.ReLU(),
            nn.Linear(size, model_dim)
        )


class BapnaAdaptered(nn.Module):
    def __init__(self, layer):
        super().__init__()
        self.adapter = BapnaAdapter()
        self.layer = layer

    def forward(self, *x):

        adap_out = self.adapter.forward(*x)
        out = self.layer(adap_out[0].unsqueeze(0))

        return out




class Model_with_adapter_bapna():

    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = ''):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Bapna Adapter'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")

        for name, params in self.model.named_parameters():
            # Freeze model except the layer norm layers
            if 'ln' not in name:
                params.requires_grad = False
        for i in range(1, 12):
            # Insert the Bapna adapter at the end of each transformer block
            # (applying it to ln_1 layer means applying it at the end of the previous transformer block)
            self.model.transformer.h[i].ln_1 = BapnaAdaptered(self.model.transformer.h[i].ln_1)
        self.model.transformer.ln_f = BapnaAdaptered(self.model.transformer.ln_f)


    def get_model(self):

        return self.model


from transformers import GPT2LMHeadModel
import torch.nn as nn

class Adapter(nn.Module):
    """
    The adapter first projects the original d-dimensional features into a 
    smaller dimension m, applies a nonlinearity, then projects back to d dimensions.
    For more information please refer to the following paper:
    https://arxiv.org/pdf/1902.00751.pdf
    """
    def __init__(self, size = 6, model_dim = 768, scale_factor = None, add_res = True):
        super().__init__()

        self.adapter_block = nn.Sequential(
            nn.Linear(model_dim, size),
            nn.ReLU(),
            nn.Linear(size, model_dim)
        )

        self.scale_factor = scale_factor
        self.add_res = add_res

    def forward(self, x):

        out = self.adapter_block(x)
        # Skip connection
        if self.add_res:
            out += x
        if self.scale_factor is not None:
            out *= self.scale_factor
        

        return out


class Adaptered(nn.Module):
    def __init__(self, orig_layer):
        super().__init__()
        self.orig_layer = orig_layer
        self.adapter = Adapter()

    def forward(self, *x):

        orig_out = self.orig_layer(*x)
        output = (self.adapter.forward(orig_out[0].unsqueeze(0))[0],)[0]

        return output




class Model_with_adapter():

    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = ''):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Houlsby Adapter'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for name, params in self.model.named_parameters():
            # Freeze model except the layer norm layers
            if 'ln' not in name:
                params.requires_grad = False
        # Insert adapters at the end of the attention and mlp sublayers
        for i in range(12):
            self.model.transformer.h[i].attn.c_proj = Adaptered(self.model.transformer.h[i].attn.c_proj)
            self.model.transformer.h[i].mlp.c_proj = Adaptered(self.model.transformer.h[i].mlp.c_proj)


    def get_model(self):

        return self.model


import torch
from torch.utils.data import DataLoader
from torch.optim import AdamW
import random
from transformers import get_linear_schedule_with_warmup
from transformers import GPT2LMHeadModel, AutoConfig
from ...fine_tuning.utils import *
from ...fine_tuning.stlr import STLR
from ...data.dataset import TL_Dataset



def load_dataloaders(base_model = 'GPT2', model_size = '', dataset_path = None, block_size = 1000, train_split = 0.95, test_split = 0.25, batch_size = 1, num_workers = 0):

    full_dataset = TL_Dataset(base_model, model_size, dataset_path, block_size)
    train_size = int(train_split * len(full_dataset))
    valtest_size = len(full_dataset) - train_size
    train_dataset, valtest_dataset = torch.utils.data.random_split(full_dataset, [train_size, valtest_size])
    test_size = int(test_split * valtest_size)
    val_size = valtest_size - test_size
    val_dataset, test_dataset = torch.utils.data.random_split(valtest_dataset, [val_size, test_size])
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True,
                              num_workers=num_workers, pin_memory=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=True,
                            num_workers=num_workers, pin_memory=True)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True,
                            num_workers=num_workers, pin_memory=True)

    return train_loader, val_loader, test_loader



def get_optimizer(model, model_obj, base_lr, weight_decay):

    if model_obj.technique == 'ULMFiT':
        model_params = apply_discriminative_lr(model, base_lr, model_obj.dlr_factor)
        optimizer = AdamW(model_params, weight_decay=weight_decay)
    else:
        optimizer = AdamW(model.parameters(), weight_decay=weight_decay, lr=base_lr) 

    return optimizer



def print_model(model, print_params=False, print_sum_weights=True, check=False):
    for idx, (name, param) in enumerate(model.named_parameters()):
        print(f'{idx}: {name}')
        if print_params:
            print(param)
        print("Requires grad:", param.requires_grad)
        print("Shape:", param.size())
        print("Sum of weights:", param.abs().sum())
        if check:
            input()
    input()


def print_lr(optimizer):
    print('Learning rates:')
    for param_group in optimizer.param_groups:
        print(param_group['lr'])
    input()


def prepend_tokens(tokens, num_tokens, token, device): 

    to_prepend = torch.full((1, num_tokens), token).to(device)
    tokens = torch.cat((to_prepend, tokens), dim=1).type(torch.long)

    return tokens


# def adjust_model(model, model_obj, seq_count):
#     """
#     Checks args and makes modifications to model as necessary.
#     This function is called before the training loop.
#     """
#     technique = model_obj.technique
#     elif technique == 'Chain Thaw':
#         model = freeze_model(model) 
#         model = chain_thaw(model, 1)

#     return model


def batch_routine(model, model_obj):

    """
    Checks args and makes modifications to model as necessary.
    This function is called in the training loop at the end of each batch.
    """
    if model_obj.technique == 'Adapter Drop Robust':
        model_obj.model = model
        model_obj.reset_adapters()
        num_drop = random.randint(0, 11)
        model_obj.drop_adapters(num_drop)
        model = model_obj.model

    return model


def process_tokens(tokens, device, technique):

    if technique == 'Prefix Tuning':
        tokens = prepend_tokens(tokens, 2, 0, device)
    if technique == 'Prompt Tuning':
        tokens = prepend_tokens(tokens, 2, 0, device)

    return tokens



def modify(model, model_obj, seq_count):

    """
    Checks args and makes modifications to model as necessary.
    This function is called in the training loop at the end of each iteration.
    """
    technique = model_obj.technique
    if technique == 'ULMFiT' or technique == 'Gradual Unfreezing':
        if seq_count % model_obj.unfreeze_freq == 0:
            model = gradually_unfreeze(model, 1)
    elif technique == 'Chain Thaw':
        if seq_count % args.unfreeze_freq == 0:
            model = chain_thaw(model, 1)

    return model



def select_scheduler(scheduler, technique, optimizer, train_size, epochs, actual_batch_size, warmup_steps):

    if technique == 'ULMFiT' or scheduler == 'STLR':
        num_iters = epochs * train_size / actual_batch_size
        scheduler = STLR(optimizer, num_iters=num_iters, cut_frac=0.1, ratio=32)
    else:
        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=warmup_steps, num_training_steps=-1
        )

    return scheduler



def get_GPT2LMH(model_size, dropout):

    config = AutoConfig.from_pretrained('gpt2' + model_size)
    if dropout is not None:
        config.resid_pdrop = dropout
        config.embd_pdrop = dropout
        config.attn_pdrop = dropout
    model = GPT2LMHeadModel.from_pretrained(pretrained_model_name_or_path = 'gpt2' + model_size, config = config)

    return model



def verify_args(args):

    """
    Verify whether the args provided are mutually consistent or not.
    """
    if args.gradual_unfreezing and args.chain_thaw:
        raise Exception("Gradual unfreezing and chain thaw cannot both be enabled together")





from transformers import GPT2LMHeadModel
import torch.nn as nn
import torch



class PromptTuned(nn.Module):

    def __init__(self, inp_layer, prompt_len):
        super().__init__()
        self.inp_layer = inp_layer
        self.prompt_len = prompt_len
        self.prompt = torch.nn.parameter.Parameter(data=torch.rand(1, prompt_len, 768))
        
        
    def forward(self, *x):

        num_inputs = x[0].size(dim=1)
        x = list(x)
        x[0] = (x[0][:, self.prompt_len:])
        x = tuple(x)
        inp_out = self.inp_layer(*x)
        cat_out = torch.cat((self.prompt, inp_out), dim=1)

        return cat_out



class Model_with_prompt_tuning():
    """
    Prompt tuning prepends a parameterised prompt to the input of the model.
    For more information, please refer to the following paper:
    https://aclanthology.org/2021.emnlp-main.243.pdf
    """
    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', prompt_len=2):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Prompt Tuning'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for params in self.model.parameters():
            params.requires_grad = False

        self.model.transformer.wte = PromptTuned(self.model.transformer.wte, prompt_len)
        


    def get_model(self):

        return self.model


from transformers import GPT2LMHeadModel
import torch.nn as nn
import torch.autograd



class STEFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input):
        return (input > 0.5).float()

    @staticmethod
    def backward(ctx, grad_output):
        return nn.functional.hardtanh(grad_output)


class StraightThroughEstimator(nn.Module):
    def __init__(self, threshold):
        super(StraightThroughEstimator, self).__init__()
        self.threshold = threshold

    def forward(self, x):
        x = STEFunction.apply(x)
        return x


class Masked(nn.Module):

    def __init__(self, orig_layer, threshold):
        super().__init__()
        self.orig_layer = orig_layer
        self.threshold = threshold
        print(self.orig_layer.weight.size())
        #input()
        self.pre_mask = nn.Parameter(torch.rand(self.orig_layer.weight.size(), requires_grad=True)).to('cuda')
        self.masking_layer = StraightThroughEstimator(self.threshold)

    def forward(self, *x):

        #orig_out = self.orig_layer(*x)
        print(self.pre_mask)
        input("Pre mask")
        mask = nn.Parameter(self.masking_layer.forward(self.pre_mask)).to('cuda')
        masked_layer = self.orig_layer
        print(self.orig_layer.weight.is_cuda)
        print(masked_layer.weight.is_cuda)
        print(mask.is_cuda)
        masked_layer.weight = nn.Parameter(self.orig_layer.weight * mask).to('cuda')
        # print("Mask:", mask)
        # input()
        output = masked_layer(*x)
        #output = (self.adapter.forward(orig_out[0].unsqueeze(0))[0],)
        return output



class Model_with_mask(nn.Module):

    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_type = ''):
        super().__init__()
        self.threshold = 0.5
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_type, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for i in range(12):
            self.model.transformer.h[i].attn.c_attn = Masked(self.model.transformer.h[i].attn.c_attn, self.threshold)
            #self.model.transformer.h[i].attn.c_proj = Masked(self.model.transformer.h[i].attn.c_proj, self.threshold)
            #self.model.transformer.h[i].mlp.c_fc = Masked(self.model.transformer.h[i].mlp.c_fc, self.threshold)
            #self.model.transformer.h[i].mlp.c_proj = Masked(self.model.transformer.h[i].mlp.c_proj, self.threshold)


    def get_model(self):

        return self.model








from transformers import GPT2LMHeadModel
import torch.nn as nn
from utils import freeze_except_bias

class Model_with_BitFit():

    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', freeze_initial_layers = 0, freeze_layer = None):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'BitFit'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
            
        self.model = freeze_except_bias(self.model)



from transformers import GPT2LMHeadModel
import torch.nn as nn


class LoRABlock(nn.Module):

    def __init__(self, r, in_dim, out_dim, alpha):

        self.r = r
        self.alpha = alpha
        super().__init__()
        self.lora_A = nn.Linear(in_dim, r)
        self.lora_B = nn.Linear(r, out_dim)
        self.lora_B.weight.data.zero_()
        self.lora_B.bias.data.zero_()

    def forward(self, x):

        out_A = self.lora_A(x)
        out = self.lora_B(out_A)

        return out

    def get_r():

        return self.r

    def get_alpha():

        return self.alpha


class LoRALayer(nn.Module):

    def __init__(self, orig_layer):
        super().__init__()
        self.orig_layer = orig_layer
        self.lora = LoRABlock(2, 768, 768, 1)

    def forward(self, *x):

        orig_out = self.orig_layer(*x)
        lora_out = self.lora.forward(*x)
        lora_out *= self.lora.get_r() / self.lora.get_alpha()

        out = orig_out + lora_out

        return out



class Model_with_LoRA():
    """
    LoRA approach freezes the pretrained model weights and injects trainable 
    rank decomposition matrices into each layer of the Transformer architecture.
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2106.09685.pdf
    """
    def __init__(self, base_model = 'GPT2', model_size = ''):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'LoRA'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for params in self.model.parameters():
            params.requires_grad = False
        for i in range(12):
            # Insert LoRA modules into the attention sublayers
            self.model.transformer.h[i].attn.c_attn = LoRALayer(self.model.transformer.h[i].attn.c_attn)
            self.model.transformer.h[i].attn.c_proj = LoRALayer(self.model.transformer.h[i].attn.c_proj)

        
    def get_model(self):

        return self.model



from ...parameter_efficient.mam_adapter import Model_with_mam_adapter
from ...parameter_efficient.prefix_tuning import Model_with_prefix_tuning
from ...parameter_efficient.prompt_tuning import Model_with_prompt_tuning
from ...parameter_efficient.parallel_adapter import Model_with_parallel_adapter
from ...parameter_efficient.adapterdrop import *
from ...parameter_efficient.adapter import Model_with_adapter
from ...parameter_efficient.adapter_bapna import Model_with_adapter_bapna
from ...parameter_efficient.lora import Model_with_LoRA
from utils import get_GPT2LMH


def select_model(base_model, model_size, technique, dropout):

    models = {
      'GPT2LMH': get_GPT2LMH(model_size, dropout),
      'Houlsby Adapter': Model_with_adapter(),
      'Bapna Adapter': Model_with_adapter_bapna(),
      'AdapterDrop Specialised': Model_with_adapterdrop_spec(),
      'AdapterDrop Robust': Model_with_adapterdrop_rob(),
      'Parallel Adapter': Model_with_parallel_adapter(),
      'MAM Adapter': Model_with_mam_adapter(),
      'LoRA': Model_with_LoRA(),
      'Prompt Tuning': Model_with_prompt_tuning(),
      'Prefix Tuning': Model_with_prefix_tuning()
    }

    model = models[technique]

    return model





from transformers import GPT2Tokenizer, GPT2LMHeadModel
import torch
import os
import warnings
from nltk.translate.bleu_score import sentence_bleu
from torchmetrics.text.rouge import ROUGEScore
from torchmetrics import SacreBLEUScore



def generate(prompt, model_obj, device, model_save_name, gen_max_len):
    """
    Feel free to modify this function as per your requirements.
    """

    if model_obj.base_model == 'GPT2':
        tokenizer = GPT2Tokenizer.from_pretrained('gpt2' + model_obj.model_size)
    else:
        raise Exception("Error: this library supports only GPT2 as the base model for now.")

    if not os.path.exists(model_save_name):
        warnings.warn("User provided model doesn't exist. Using default model instead.")
    else:
        state_dict = torch.load(model_save_name)
        model.load_state_dict(state_dict)

    model = model_obj.model
    model = model.to(device)
    input_ids = tokenizer.encode(prompt, return_tensors='pt').to(device)
    outputs_ = model.generate(
        input_ids,
        max_length=gen_max_len,
        do_sample=True,
        top_k=20,
        no_repeat_ngram_size=2,
        temperature=0.7
        )

    output = tokenizer.decode(outputs_[0], skip_special_tokens=True)

    return output



def evaluate(test_loader, model_obj, test_split_token, test_end_token, model_save_name, gen_max_len):
    """
    Calculate the BLEU, ROUGE, and SacreBLEU scores for the model.
    Returns a dictionary of metric scores.
    """
    print("EVALUATING" + "." * 20)

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    BLEUscores = []
    ROUGEscores = []
    SacreBLEUscores = []

    rouge = ROUGEScore()
    sacre_bleu = SacreBLEUScore()

    if model_obj.base_model == 'GPT2':
        tokenizer = GPT2Tokenizer.from_pretrained('gpt2' + model_obj.model_size)
    else:
        raise Exception("Error: this library supports only GPT2 as the base model for now.")
        
    for text in test_loader:

        text = tokenizer.decode(text[0], skip_special_tokens=True)
        if test_split_token is None:
            inp_num_words = 10
            text = text.split(" ")
            inp = " ".join(text[:inp_num_words])
            ref = text[inp_num_words:]
        else:
            text = text.split(test_split_token)
            inp = text[0]
            ref = inp + text[1].replace(test_end_token, '')

        gen = generate(inp, model_obj, device, model_save_name, gen_max_len)

        if test_end_token is None:
            ref = inp + " " + " ".join(ref[:gen_max_len])
        else:
            gen = gen.split(test_end_token)[0]
            gen = gen.replace(test_split_token, '')

        ROUGEscores.append(rouge(gen, ref)['rouge1_recall'].item())
        SacreBLEUscores.append(sacre_bleu([gen], [[ref]]).item())
        ref = [ref.split(" ")]
        gen = gen.split(" ")
        BLEUscores.append(sentence_bleu(ref, gen))

    BLEUscore = sum(BLEUscores) / len(BLEUscores)
    ROUGEscore = sum(ROUGEscores) / len(ROUGEscores)
    SacreBLEUscore = sum(SacreBLEUscores) / len(SacreBLEUscores)

    metrics = {
      'BLEU': BLEUscore,
      'ROUGE (recall)': ROUGEscore,
      'SacreBLEU': SacreBLEUscore
    }

    return metrics











from transformers import GPT2LMHeadModel
import torch.nn as nn
from adapter import Adaptered


class Model_with_adapterdrop_spec():
    """
    Specialized AdapterDrop: Removing adapters from the first n transformer 
    layers, where n is fixed during training. This yields separate models for
    each possible n.
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2010.11918.pdf
    """
    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', num_drop = 0):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'AdapterDrop Specialised'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")        
        self.num_drop = num_drop
        for params in self.model.parameters():
            params.requires_grad = False
        for i in range(self.num_drop, 12):
            self.model.transformer.h[i].attn.c_proj = Adaptered(self.model.transformer.h[i].attn.c_proj)
            self.model.transformer.h[i].mlp.c_proj = Adaptered(self.model.transformer.h[i].mlp.c_proj)


    def get_model(self):

        return self.model

    def get_num_drop(self):

        return self.num_drop


class RobustAdaptered(Adaptered):

    def __init__(self, orig_layer, drop=False):

        super(RobustAdaptered, self).__init__(orig_layer)
        self.drop = drop

    def forward(self, *x):

        orig_out = self.orig_layer(*x)
        if self.drop:
            return orig_out 
        output = (self.adapter.forward(orig_out[0].unsqueeze(0))[0],)[0]

        return output

    

class Model_with_adapterdrop_rob():
    """
    Robust AdapterDrop: Drawing the integer n randomly from [0, 11] for each 
    training batch. This yields one robust model that is applicable to a varying 
    number of dropped layers.
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2010.11918.pdf
    """
    def __init__(self, base_model = 'GPT2', model_size = ''):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'AdapterDrop Robust'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")        
        for params in self.model.parameters():
            params.requires_grad = False
        for i in range(12):
            self.model.transformer.h[i].attn.c_proj = RobustAdaptered(self.model.transformer.h[i].attn.c_proj)
            self.model.transformer.h[i].mlp.c_proj = RobustAdaptered(self.model.transformer.h[i].mlp.c_proj)


    def get_model(self):

        return self.model

    def drop_adapters(self, n = 0):

        for i in range(n):
            self.model.transformer.h[i].attn.c_proj.drop = True
            self.model.transformer.h[i].mlp.c_proj.drop = True

    def reset_adapters(self):

        for i in range(12):
            self.model.transformer.h[i].attn.c_proj.drop = False
            self.model.transformer.h[i].mlp.c_proj.drop = False



from transformers import GPT2LMHeadModel
import torch.nn as nn
from utils import freeze_model, gradually_unfreeze

class Model_with_gradual_unfreezing():

    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', unfreeze_freq = 100):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Gradual Unfreezing'
        self.unfreeze_freq = unfreeze_freq
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")

        self.model = freeze_model(self.model)
        self.model = gradually_unfreeze(self.model, 1)


class Model_with_ULMFiT(Model_with_gradual_unfreezing):

    def __init__(self, base_model = 'GPT2', model_size = '', unfreeze_freq = 100, dlr_factor = 0.384615):
        super(Model_with_ULMFiT, self).__init__(base_model, model_size, unfreeze_freq)
        self.technique = 'ULMFiT'
        self.dlr_factor = dlr_factor


from torch.optim.lr_scheduler import _LRScheduler
import warnings
import math

class STLR(_LRScheduler):

    """
    SLANTED TRIANGULAR LEARNING RATE
    Linearly increases the learning rate first and then linearly decays it 
    according to a schedule specified in the following paper:
    https://arxiv.org/pdf/1801.06146.pdf
    """

    def __init__(self, optimizer, last_epoch=-1, verbose=False, num_iters=0, cut_frac=0, ratio=0):

        self.num_iters = num_iters
        self.cut_frac = cut_frac
        self.ratio = ratio
        self.cut = math.floor(num_iters * self.cut_frac)
        super(STLR, self).__init__(optimizer, last_epoch, verbose)


    def get_lr(self):

        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        if self._step_count < self.cut:
            p = self._step_count / self.cut
        else:
            p = 1 - ((self._step_count - self.cut)/(self.cut*(1/self.cut_frac - 1)))

        factor = (1 + p*(self.ratio - 1))/self.ratio

        return [max_lr * factor
                for max_lr in self.base_lrs]












from transformers import GPT2LMHeadModel
import torch.nn as nn
import torch


# To figure out how to make this local
to_add = None

class PrefixTuned(nn.Module):

    def __init__(self, layer, prefix_len, k, replace_prefix, calc_dif):
        super().__init__()
        self.layer = layer
        self.prefix_len = prefix_len
        self.calc_dif = calc_dif
        self.replace_prefix = replace_prefix
        self.prefix_mlp = nn.Linear(k, 768)
        self.prefix_prime = nn.parameter.Parameter(torch.rand(prefix_len, k))
        self.prefix = None
        self.old_inp = None
        self.new_inp = None
        
        
    def forward(self, *x):

        global to_add
        # Parameterisation of the prefix
        self.prefix = self.prefix_mlp(self.prefix_prime).unsqueeze(0)
        self.old_inp = x
        num_inputs = x[0].size(dim=1)
        x = list(x)
        if self.replace_prefix:
            x[0] = torch.cat((self.prefix, x[0][:, self.prefix_len:]), dim=1)
        else:
            x[0] = torch.cat((self.prefix, x[0]), dim=1)
        x = tuple(x)
        self.prefixed = True
        out = self.layer(x[0])
        self.new_inp = x
        if self.calc_dif:
            to_add = torch.sub(self.new_inp[0], self.old_inp[0])

        return out



class SumOutput(nn.Module):

      def __init__(self, layer):
          super().__init__()
          self.layer = layer

      def forward(self, *x):

          out = self.layer(*x)
          out += to_add

          return out




class Model_with_prefix_tuning():
    """
    Prefix Tuning keeps language model parameters frozen, but optimizes a small 
    continuous task-specific vector (called the prefix).
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2101.00190.pdf
    """
    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', prefix_len = 10, k = 42):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Prefix Tuning'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for params in self.model.parameters():
            params.requires_grad = False
        for i in range(1, 12):
              self.model.transformer.h[i].ln_1 = PrefixTuned(self.model.transformer.h[i].ln_1, prefix_len, k, True, True)
              self.model.transformer.h[i].attn.resid_dropout = SumOutput(self.model.transformer.h[i].attn.resid_dropout)
              self.model.transformer.h[i].mlp.dropout = SumOutput(self.model.transformer.h[i].mlp.dropout)
        self.model.transformer.ln_f = PrefixTuned(self.model.transformer.ln_f, prefix_len, k, True, True)


    def get_model(self):

        return self.model


from transformers import GPT2LMHeadModel
import torch.nn as nn
from parallel_adapter import LayerAdaptered, Sum
from prefix_tuning import PrefixTuned



class PrefixRemoved(nn.Module):

    def __init__(self, layer, prefix_len):
        super().__init__()
        self.layer = layer
        self.prefix_len = prefix_len

    def forward(self, *x):

        out = self.layer(*x)
        out = out[:, self.prefix_len:, :]

        return out



class Model_with_mam_adapter():

    """
    An adapter variant which is designed by integrating favorable design 
    elements from other popular parametre efficient approaches like adapters, 
    prefix tuning and LoRA.
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2110.04366.pdf
    """
    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', prefix_len = 10, k = 42, r = 512, scale = 4):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = "MAM Adapter"
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for params in self.model.parameters():
            params.requires_grad = False
        for i in range(12):
            self.model.transformer.h[i].attn.c_attn = PrefixTuned(self.model.transformer.h[i].attn.c_attn, prefix_len, k, False, False)
            self.model.transformer.h[i].attn.resid_dropout = PrefixRemoved(self.model.transformer.h[i].attn.resid_dropout, prefix_len)
            self.model.transformer.h[i].mlp.c_fc = LayerAdaptered(self.model.transformer.h[i].mlp.c_fc, r, scale)
            self.model.transformer.h[i].mlp.dropout = Sum(self.model.transformer.h[i].mlp.dropout, self.model.transformer.h[i].mlp.c_fc)

    def get_model(self):

        return self.model




def apply_discriminative_lr(model, base_lr, dlr_factor):
    """
    Applies learning rate to parameter groups in an exponentially decreasing 
    manner. Refer to the following paper for more information:
    https://arxiv.org/pdf/1801.06146.pdf
    """

    lr = base_lr
    parameters = []
    prev_group_name = ''

    for name, param in reversed(list(model.named_parameters())):
        cur_group_name = get_group_name(name)
        if cur_group_name != prev_group_name:
            if prev_group_name != '':
                lr *= dlr_factor
            prev_group_name = cur_group_name

        parameters += [{'params': param, 'lr': lr}]

    parameters.reverse()

    return parameters



def freeze_model(model):

    for p in model.parameters():
        p.requires_grad = False

    return model


def freeze_except_bias(model):

    for name, params in model.named_parameters():
        if 'bias' in name:
            params.requires_grad = True
        else:
            params.requires_grad = False

    return model


def get_group_name(name):
    """
    Returns the group name given a model layer name.
    Note: this is specific to gpt-2 for the moment.
    """
    tags = name.split('.')
    if len(tags) == 3:
        return ".".join(tags[:2])
    else:
        return ".".join(tags[:4])


def last_unfrozen_group(model):
    """
    Returns the name of the last unfrozen group and of the previous-to-last.
    """
    prev_group_name = ''
    for name, param in model.named_parameters():
        group_name = get_group_name(name)
        if not param.requires_grad:
            if group_name != prev_group_name:
                prev_group_name = group_name
        else:
            break

    return group_name, prev_group_name


def get_prev_group_name(model, group_name):
    """
    Given a group name, return the name of the previous group.
    """
    prev_group_name = ''
    for name, param in model.named_parameters():
        cur_group_name = get_group_name(name)
        if cur_group_name == group_name:
            return prev_group_name
        else:
            if cur_group_name != prev_group_name:
                prev_group_name = cur_group_name


def get_next_group_name(model, group_name):
    """
    Given a group name, return the name of the next group.
    """
    next_group_name = ''
    for name, param in reversed(list(model.named_parameters())):
        cur_group_name = get_group_name(name)
        if cur_group_name == group_name:
            return next_group_name
        else:
            if cur_group_name != next_group_name:
                next_group_name = cur_group_name


def gradually_unfreeze(model, num_groups):
    """
    Unfreezes the model layer by layer from the top according to 
    args.unfreeze_freq. Refer to the following paper for more information:
    https://arxiv.org/pdf/1801.06146.pdf
    """
  
    groups_to_unfreeze = []

    last_unfrozen, to_unfreeze = last_unfrozen_group(model)
    groups_to_unfreeze.append(to_unfreeze)
    for i in range(num_groups - 1):
        group_name = get_prev_group_name(groups_to_unfreeze[-1])
        groups_to_unfreeze.append(group_name)

    for name, param in model.named_parameters():
        for group in groups_to_unfreeze:
            if group in name:
                param.requires_grad = True
                break

    return model


def chain_thaw(model, num_groups):
    """
    Unfreeze layers from the top one by one and one at a time.
    """

    groups_to_freeze = []
    groups_to_unfreeze = []
    last_unfrozen, to_unfreeze = last_unfrozen_group(model)
    groups_to_freeze.append(last_unfrozen)
    groups_to_unfreeze.append(to_unfreeze)

    for i in range(num_groups - 1):
        groups_to_freeze.append(get_next_group_name(model, groups_to_freeze[-1]))
        groups_to_unfreeze.append(get_prev_group_name(model, groups_to_unfreeze[-1]))

    for name, param in model.named_parameters():
        for group in groups_to_unfreeze:
            if group in name:
                param.requires_grad = True
                break
        else:
            for group in groups_to_freeze:
                if group in name:
                    param.requires_grad = False
                    break

    return model



def unfreeze_last_n_layers(model, n):
    """
    Unfreeze last n layers (groups) of the model.
    """
    prev_group_name = ''
    unfrozen = 0
    for name, param in reversed(list(model.named_parameters())):
        group_name = get_group_name(name)
        if group_name != prev_group_name:
            prev_group_name = group_name
            unfrozen += 1
            if unfrozen > n:
                break
        param.requires_grad = True

    return model


def freeze_first_n_layers(model, n):
    """
    Freeze first n layers (groups) of the model
    """
    prev_group_name = ''
    frozen = 0
    for name, param in model.named_parameters():
        group_name = get_group_name(name)
        if group_name != prev_group_name:
            prev_group_name = group_name
            frozen += 1
            if frozen > n:
                break
        param.requires_grad = False

    return model



def freeze_nth_layer(model, n):
    """
    Freeze nth layer (group) of the model
    """
    prev_group_name = ''
    layer = 0
    for name, param in model.named_parameters():
        group_name = get_group_name(name)
        if group_name != prev_group_name:
            prev_group_name = group_name
            layer += 1
            if layer > n:
                break
        if layer == n:
            param.requires_grad = False

    return model





from transformers import GPT2LMHeadModel
import torch.nn as nn
from adapter import Adapter



class LayerAdaptered(nn.Module):
    """
    Inserts an adapter parallel to orig_layer
    """
    def __init__(self, orig_layer, adap_size, scale_factor):
        super().__init__()
        self.orig_layer = orig_layer
        self.adapter = Adapter(size=adap_size, add_res=False, scale_factor=scale_factor)
        self.adapter_out = 0

    def forward(self, *x):

        orig_out = self.orig_layer(*x)
        self.adapter_out = self.adapter.forward(*x)

        return orig_out

    def get_adap_out(self):

        return self.adapter_out




class Sum(nn.Module):

    def __init__(self, orig_layer, adap_obj):
        super().__init__()
        self.orig_layer = orig_layer
        self.adap_obj = adap_obj

    def forward(self, *x):
        
        orig_out = self.orig_layer(*x)
        out = orig_out + self.adap_obj.adapter_out

        return out
	  




class EmbeddingAdaptered(nn.Module):
  
    def __init__(self, embed_layer):
        super().__init__()
        self.embed_layer = embed_layer
        self.adapter = Adapter()

    def forward(self, *x):

        embed_out = self.embed_layer(*x)
        adapter_out = (self.adapter.forward(embed_out[0].unsqueeze(0))[0],)[0]
        out = embed_out + adapter_out

        return out




class Model_with_parallel_adapter():
    """
    An adapter variant that arranges adapters in parallel instead of in series.
    For more information, please refer to the following paper:
    https://arxiv.org/pdf/2104.08154v1.pdf
    """
    # model_type options: '', '-medium', '-large'
    def __init__(self, base_model = 'GPT2', model_size = '', adap_size = 100):
        self.base_model = base_model
        self.model_size = model_size
        self.technique = 'Parallel Adapter'
        if base_model == 'GPT2':
            self.model = GPT2LMHeadModel.from_pretrained('gpt2' + model_size, return_dict=False)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        for params in self.model.parameters():
            params.requires_grad = False
        self.model.transformer.wte = EmbeddingAdaptered(self.model.transformer.wte)
        for i in range(12):
            self.model.transformer.h[i].attn.c_attn = LayerAdaptered(self.model.transformer.h[i].attn.c_attn, adap_size, None)
            self.model.transformer.h[i].attn.resid_dropout = Sum(self.model.transformer.h[i].attn.resid_dropout, self.model.transformer.h[i].attn.c_attn)
            self.model.transformer.h[i].mlp.c_fc = LayerAdaptered(self.model.transformer.h[i].mlp.c_fc, adap_size, None)
            self.model.transformer.h[i].mlp.dropout = Sum(self.model.transformer.h[i].mlp.dropout, self.model.transformer.h[i].mlp.c_fc)

    def get_model(self):

        return self.model


from torch.utils.data import Dataset
from transformers import GPT2Tokenizer
import torch
#from create_dataset import generate_dataset


class TL_Dataset(Dataset):

    def __init__(self, base_model, model_size, dataset_path, block_size):
        """
        Reads data from a txt file and chunks it into args.block_size batches.
        Feel free to modify this file as per your requirements.
        """

        super().__init__()
        #generate_dataset(MAX_LINE_LEN)
        if base_model == 'GPT2':
            self.tokenizer = GPT2Tokenizer.from_pretrained('gpt2' + model_size)
        else:
            raise Exception("Error: this library supports only GPT2 as the base model for now.")
        self.dataset_path = dataset_path
        self.sentences = []
        self.text = ''

        with open(dataset_path) as file:

            for line in file:
                self.text += line.strip('\n')

            tokenized_text = self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(self.text))

            for i in range(0, len(tokenized_text) - block_size + 1, block_size):  # Truncate in block of block_size
                self.sentences.append(torch.tensor(self.tokenizer.build_inputs_with_special_tokens(tokenized_text[i: i + block_size])))


    def __len__(self):
        return len(self.sentences)

    def __getitem__(self, item):
        return self.sentences[item]


import torch
from torch.utils.tensorboard import SummaryWriter
import argparse
import os
import time, datetime
from tqdm import tqdm
import transformers
import random
import sys
from utils import *



def train(
    model_obj = None,
    train_loader = None,
    val_loader = None,
    actual_batch_size = 15,
    epochs = 20,
    base_lr = 0.001,
    weight_decay = 0.001,
    scheduler = None,
    warmup_steps = 5000,
    model_save_name = None,
    save_model_freq = 300,
    val_freq = 100,
    dropout = 0.1,
    write_logs = True,
    logs_folder = 'runs',
    delete_logs = True,
    load_existing_model = True,
    ):
  

    if model_obj is None:
        raise Exception("Model object was not provided.")
    if train_loader is None or val_loader is None:
        raise Exception("train_loader or val_loader were not provided.")
    if model_save_name is None:
        raise Exception("Path to directory for saving model was not provided.")

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    base_model = model_obj.base_model
    model_size = model_obj.model_size

    if hasattr(model_obj, 'model'):
        model = model_obj.model
    else:
        model = model_obj

    if write_logs:
        writer = SummaryWriter(logs_folder)

    if delete_logs:
        # Delete all the logs
        for root, dirs, files in os.walk(logs_folder):
            for file in files:
                os.remove(os.path.join(root, file))
        print("Logs deleted")

    if load_existing_model:
        state_dict = torch.load(model_save_name)
        model.load_state_dict(state_dict)
        print("Existing model loaded")

    model = model.to(device)

    optimizer = get_optimizer(model, model_obj, base_lr, weight_decay)
    scheduler = select_scheduler(scheduler, model_obj.technique, optimizer, len(train_loader), epochs, actual_batch_size, warmup_steps)

    model.train()
    model_obj.model = model

    seq_count = 0
    start = time.time()


    for epoch in range(epochs):

        print(f"EPOCH {epoch} started" + '=' * 30)

        for train_counter, train_batch in enumerate(train_loader, 0):

            tokens = train_batch.to(device)
            tokens = process_tokens(tokens, device, model_obj.technique)
            outputs = model(tokens, labels=tokens)
            loss = outputs[0]
            loss.backward()
            seq_count += 1

            print('[%d, %5d] train loss: %.5f' % (epoch + 1, seq_count, loss.detach().data))
            if write_logs:
                writer.add_scalar("train_loss", float(loss.detach().data), seq_count)

            # Resorting to this approach as processing input sequences in 
            # parallel would often cause 'CUDA out of memory'
            if seq_count % actual_batch_size == 0:
              optimizer.step()
              scheduler.step()
              optimizer.zero_grad()
              model.zero_grad()
              model = batch_routine(model, model_obj)

            if seq_count % save_model_freq == 0:
                torch.save(model.state_dict(), model_save_name)

            if seq_count % val_freq == 0:
                validate(model, model_obj, val_loader, device, seq_count // val_freq, writer)
                print("Time elapsed:", str(datetime.timedelta(seconds=time.time() - start)))
                model.train()

            if hasattr(model_obj, 'model'):
                model_obj.model = model

            model = modify(model, model_obj, seq_count)
            

            


def validate(model, model_obj, val_loader, device, batch_count, writer):

    model.eval()
    counter = 0
    with torch.no_grad():
        running_loss = 0

        for val_counter, val_batch in enumerate(tqdm(val_loader), 0):

            tokens = val_batch.to(device)
            tokens = process_tokens(tokens, device, model_obj.technique)
            outputs = model(tokens, labels=tokens)
            loss = outputs[0]
            counter += 1
            running_loss += loss

        print('[%d       ] validation loss: %.5f' % (batch_count,
                                                     running_loss / len(val_loader)))
        writer.add_scalar('val loss', running_loss / len(val_loader),
                          batch_count)


